<div xmlns="http://www.w3.org/1999/xhtml"
     xmlns:py="http://genshi.edgewall.org/"
     xmlns:xi="http://www.w3.org/2001/XInclude">
  <xi:include href="xinclude${2}.html" />
  ${greeting('world')}
  Goodbye!
</div>
