<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"
     xmlns:py="http://genshi.edgewall.org/"
     py:strip="">
  <py:match path="//div" once="true">
    <div xmlns="http://www.w3.org/1999/xhtml">
      <span>
        ${select('*')}
      </span>
    </div>
  </py:match>
  <p py:def="greeting(name)" class="greeting">
    Hello, ${name}!
  </p>
</html>
